from __future__ import annotations

import asyncio
import json

from memphis.exceptions import MemphisError
from memphis.utils import default_error_handler, get_internal_name
from memphis.message import Message


class Consumer:
    MAX_BATCH_SIZE = 5000

    def __init__(
        self,
        connection,
        station_name: str,
        consumer_name,
        consumer_group,
        pull_interval_ms: int,
        batch_size: int,
        batch_max_time_to_wait_ms: int,
        max_ack_time_ms: int,
        max_msg_deliveries: int = 10,
        error_callback=None,
        start_consume_from_sequence: int = 1,
        last_messages: int = -1,
    ):
        self.connection = connection
        self.station_name = station_name.lower()
        self.consumer_name = consumer_name.lower()
        self.consumer_group = consumer_group.lower()
        self.pull_interval_ms = pull_interval_ms
        self.batch_size = batch_size
        self.batch_max_time_to_wait_ms = batch_max_time_to_wait_ms
        self.max_ack_time_ms = max_ack_time_ms
        self.max_msg_deliveries = max_msg_deliveries
        self.ping_consumer_invterval_ms = 30000
        if error_callback is None:
            error_callback = default_error_handler
        self.t_ping = asyncio.create_task(self.__ping_consumer(error_callback))
        self.start_consume_from_sequence = start_consume_from_sequence
        self.last_messages = last_messages
        self.context = {}
        self.dls_messages = []
        self.dls_current_index = 0
        self.dls_callback_func = None
        self.t_dls = asyncio.create_task(self.__consume_dls())
        self.t_consume = None

    def set_context(self, context):
        """Set a context (dict) that will be passed to each message handler call."""
        self.context = context

    def consume(self, callback):
        """Consume events."""
        self.dls_callback_func = callback
        self.t_consume = asyncio.create_task(self.__consume(callback))

    async def __consume(self, callback):
        subject = get_internal_name(self.station_name)
        consumer_group = get_internal_name(self.consumer_group)
        self.psub = await self.connection.broker_connection.pull_subscribe(
            subject + ".final", durable=consumer_group
        )
        while True:
            if self.connection.is_connection_active and self.pull_interval_ms:
                try:
                    memphis_messages = []
                    msgs = await self.psub.fetch(self.batch_size)
                    for msg in msgs:
                        memphis_messages.append(
                            Message(msg, self.connection, self.consumer_group)
                        )
                    await callback(memphis_messages, None, self.context)
                    await asyncio.sleep(self.pull_interval_ms / 1000)

                except asyncio.TimeoutError:
                    await callback(
                        [], MemphisError("Memphis: TimeoutError"), self.context
                    )
                    continue
                except Exception as e:
                    if self.connection.is_connection_active:
                        raise MemphisError(str(e)) from e
                    return
            else:
                break

    async def __consume_dls(self):
        subject = get_internal_name(self.station_name)
        consumer_group = get_internal_name(self.consumer_group)
        try:
            subscription_name = "$memphis_dls_" + subject + "_" + consumer_group
            self.consumer_dls = await self.connection.broker_manager.subscribe(
                subscription_name, subscription_name
            )
            async for msg in self.consumer_dls.messages:
                index_to_insert = self.dls_current_index
                if index_to_insert >= 10000:
                    index_to_insert %= 10000
                self.dls_messages.insert(
                    index_to_insert, Message(
                        msg, self.connection, self.consumer_group)
                )
                self.dls_current_index += 1
                if self.dls_callback_func != None:
                    await self.dls_callback_func(
                        [Message(msg, self.connection, self.consumer_group)],
                        None,
                        self.context,
                    )
        except Exception as e:
            if self.dls_callback_func != None:
                await self.dls_callback_func([], MemphisError(str(e)), self.context)
                return

    async def fetch(self, batch_size: int = 10):
        """Fetch a batch of messages."""
        messages = []
        if self.connection.is_connection_active:
            try:
                if batch_size > self.MAX_BATCH_SIZE:
                    raise MemphisError(
                        f"Batch size can not be greater than {self.MAX_BATCH_SIZE}")
                self.batch_size = batch_size
                if len(self.dls_messages) > 0:
                    if len(self.dls_messages) <= batch_size:
                        messages = self.dls_messages
                        self.dls_messages = []
                        self.dls_current_index = 0
                    else:
                        messages = self.dls_messages[0:batch_size]
                        del self.dls_messages[0:batch_size]
                        self.dls_current_index -= len(messages)
                    return messages

                durable_name = ""
                if self.consumer_group != "":
                    durable_name = get_internal_name(self.consumer_group)
                else:
                    durable_name = get_internal_name(self.consumer_name)
                subject = get_internal_name(self.station_name)
                self.psub = await self.connection.broker_connection.pull_subscribe(
                    subject + ".final", durable=durable_name
                )
                msgs = await self.psub.fetch(batch_size)
                for msg in msgs:
                    messages.append(
                        Message(msg, self.connection, self.consumer_group))
                return messages
            except Exception as e:
                if "timeout" not in str(e):
                    raise MemphisError(str(e)) from e
        else:
            return messages

    async def __ping_consumer(self, callback):
        while True:
            try:
                await asyncio.sleep(self.ping_consumer_invterval_ms / 1000)
                consumer_group = get_internal_name(self.consumer_group)
                await self.connection.broker_connection.consumer_info(
                    self.station_name, consumer_group, timeout=30
                )

            except Exception as e:
                callback(MemphisError(str(e)))

    async def destroy(self):
        """Destroy the consumer."""
        if self.t_consume is not None:
            self.t_consume.cancel()
        if self.t_dls is not None:
            self.t_dls.cancel()
        if self.t_ping is not None:
            self.t_ping.cancel()
        self.pull_interval_ms = None
        try:
            destroy_consumer_req = {
                "name": self.consumer_name,
                "station_name": self.station_name,
                "username": self.connection.username
            }
            consumer_name = json.dumps(
                destroy_consumer_req, indent=2).encode("utf-8")
            res = await self.connection.broker_manager.request(
                "$memphis_consumer_destructions", consumer_name, timeout=5
            )
            error = res.data.decode("utf-8")
            if error != "" and not "not exist" in error:
                raise MemphisError(error)
            self.dls_messages.clear()
            internal_station_name = get_internal_name(self.station_name)
            map_key = internal_station_name + "_" + self.consumer_name.lower()
            del self.connection.consumers_map[map_key]
        except Exception as e:
            raise MemphisError(str(e)) from e



import json

from memphis.exceptions import MemphisError
from memphis.utils import get_internal_name


class Station:
    def __init__(self, connection, name: str):
        self.connection = connection
        self.name = name.lower()

    async def destroy(self):
        """Destroy the station."""
        try:
            name_req = {"station_name": self.name, "username": self.connection.username}
            station_name = json.dumps(name_req, indent=2).encode("utf-8")
            res = await self.connection.broker_manager.request(
                "$memphis_station_destructions", station_name, timeout=5
            )
            error = res.data.decode("utf-8")
            if error != "" and not "not exist" in error:
                raise MemphisError(error)

            internal_station_name = get_internal_name(self.name)
            sub = self.connection.schema_updates_subs.get(internal_station_name)
            task = self.connection.schema_tasks.get(internal_station_name)
            if internal_station_name in self.connection.schema_updates_data:
                del self.connection.schema_updates_data[internal_station_name]
            if internal_station_name in self.connection.schema_updates_subs:
                del self.connection.schema_updates_subs[internal_station_name]
            if internal_station_name in self.connection.producers_per_station:
                del self.connection.producers_per_station[internal_station_name]
            if internal_station_name in self.connection.schema_tasks:
                del self.connection.schema_tasks[internal_station_name]
            if task is not None:
                task.cancel()
            if sub is not None:
                await sub.unsubscribe()

            self.connection.producers_map = {
                k: v
                for k, v in self.connection.producers_map.items()
                if self.name not in k
            }

            self.connection.consumers_map = {
                k: v
                for k, v in self.connection.consumers_map.items()
                if self.name not in k
            }

        except Exception as e:
            raise MemphisError(str(e)) from e



from __future__ import annotations

import json

from memphis.exceptions import MemphisConnectError, MemphisError


class Message:
    def __init__(self, message, connection, cg_name):
        self.message = message
        self.connection = connection
        self.cg_name = cg_name

    async def ack(self):
        """Ack a message is done processing."""
        try:
            await self.message.ack()
        except Exception as e:
            if (
                "$memphis_pm_id" in self.message.headers
                and "$memphis_pm_cg_name" in self.message.headers
            ):
                try:
                    msg = {
                        "id": int(self.message.headers["$memphis_pm_id"]),
                        "cg_name": self.message.headers["$memphis_pm_cg_name"],
                    }
                    msg_to_ack = json.dumps(msg).encode("utf-8")
                    await self.connection.broker_manager.publish(
                        "$memphis_pm_acks", msg_to_ack
                    )
                except Exception as e_1:
                    raise MemphisConnectError(str(e_1))
            else:
                raise MemphisConnectError(str(e)) from e
            return

    def get_data(self):
        """Receive the message."""
        try:
            return bytearray(self.message.data)
        except Exception:
            return

    def get_headers(self):
        """Receive the headers."""
        try:
            return self.message.headers
        except Exception:
            return

    def get_sequence_number(self):
        """Get message sequence number."""
        try:
            return self.message.metadata.sequence.stream
        except Exception:
            return

    async def delay(self, delay):
        """Delay and resend the message after delay seconds"""
        if (
            "$memphis_pm_id" in self.message.headers
            and "$memphis_pm_cg_name" in self.message.headers
        ):
            raise MemphisError("cannot delay DLS message")
        try:
            await self.message.nak(delay=delay)
        except Exception as e:
            raise MemphisError(str(e)) from e



from __future__ import annotations

import asyncio
import json
import time
from typing import Union

import graphql
from graphql import parse as parse_graphql
from graphql import validate as validate_graphql
from jsonschema import validate
import google.protobuf.json_format as protobuf_json_format
from memphis.exceptions import MemphisError, MemphisSchemaError
from memphis.headers import Headers
from memphis.utils import get_internal_name

schemaverse_fail_alert_type = "schema_validation_fail_alert"


class Producer:
    def __init__(
        self, connection, producer_name: str, station_name: str, real_name: str
    ):
        self.connection = connection
        self.producer_name = producer_name.lower()
        self.station_name = station_name
        self.internal_station_name = get_internal_name(self.station_name)
        self.loop = asyncio.get_running_loop()
        self.real_name = real_name

    async def validate_msg(self, message):
        if self.connection.schema_updates_data[self.internal_station_name] != {}:
            schema_type = self.connection.schema_updates_data[
                self.internal_station_name
            ]["type"]
            if schema_type == "protobuf":
                message = self.validate_protobuf(message)
                return message
            if schema_type == "json":
                message = self.validate_json_schema(message)
                return message
            if schema_type == "graphql":
                message = self.validate_graphql(message)
                return message
            if hasattr(message, "SerializeToString"):
                msg_to_send = message.SerializeToString()
                return msg_to_send
        elif isinstance(message, str):
            message = message.encode("utf-8")
            return message
        elif isinstance(message, graphql.language.ast.DocumentNode):
            msg = message
            message = str(msg.loc.source.body)
            message = message.encode("utf-8")
            return message
        elif hasattr(message, "SerializeToString"):
            msg_to_send = message.SerializeToString()
            return msg_to_send
        elif not isinstance(message, bytearray) and not isinstance(message, dict):
            raise MemphisSchemaError("Unsupported message type")
        else:
            if isinstance(message, dict):
                message = bytearray(json.dumps(message).encode("utf-8"))
            return message

    def validate_protobuf(self, message):
        proto_msg = self.connection.proto_msgs[self.internal_station_name]
        msg_to_send = ""
        try:
            if isinstance(message, bytearray):
                msg_to_send = bytes(message)
                try:
                    proto_msg.ParseFromString(msg_to_send)
                    proto_msg.SerializeToString()
                    msg_to_send = msg_to_send.decode("utf-8")
                except Exception as e:
                    if "parsing message" in str(e):
                        e = "Invalid message format, expecting protobuf"
                    raise MemphisSchemaError(str(e))
                return message
            if hasattr(message, "SerializeToString"):
                msg_to_send = message.SerializeToString()
                proto_msg.ParseFromString(msg_to_send)
                proto_msg.SerializeToString()
                try:
                    proto_msg.ParseFromString(msg_to_send)
                    proto_msg.SerializeToString()
                except Exception as e:
                    if "parsing message" in str(e):
                        e = "Error parsing protobuf message"
                    raise MemphisSchemaError(str(e))
                return msg_to_send
            elif isinstance(message, dict):
                try:
                    protobuf_json_format.ParseDict(message, proto_msg)
                    msg_to_send = proto_msg.SerializeToString()
                    return msg_to_send
                except Exception as e:
                    raise MemphisSchemaError(str(e))
            else:
                raise MemphisSchemaError("Unsupported message type")

        except Exception as e:
            raise MemphisSchemaError("Schema validation has failed: " + str(e))

    def validate_json_schema(self, message):
        try:
            if isinstance(message, bytearray):
                try:
                    message_obj = json.loads(message)
                except Exception as e:
                    raise Exception("Expecting Json format: " + str(e))
            elif isinstance(message, dict):
                message_obj = message
                message = bytearray(json.dumps(message_obj).encode("utf-8"))
            else:
                raise Exception("Unsupported message type")

            validate(
                instance=message_obj,
                schema=self.connection.json_schemas[self.internal_station_name],
            )
            return message
        except Exception as e:
            raise MemphisSchemaError("Schema validation has failed: " + str(e))

    def validate_graphql(self, message):
        try:
            if isinstance(message, bytearray):
                msg = message.decode("utf-8")
                msg = parse_graphql(msg)
            elif isinstance(message, str):
                msg = parse_graphql(message)
                message = message.encode("utf-8")
            elif isinstance(message, graphql.language.ast.DocumentNode):
                msg = message
                message = str(msg.loc.source.body)
                message = message.encode("utf-8")
            else:
                raise MemphisError("Unsupported message type")
            validate_res = validate_graphql(
                schema=self.connection.graphql_schemas[self.internal_station_name],
                document_ast=msg,
            )
            if len(validate_res) > 0:
                raise Exception(
                    "Schema validation has failed: " + str(validate_res))
            return message
        except Exception as e:
            if "Syntax Error" in str(e):
                e = "Invalid message format, expected GraphQL"
            raise Exception("Schema validation has failed: " + str(e))

    async def produce(
        self,
        message,
        ack_wait_sec: int = 15,
        headers: Union[Headers, None] = None,
        async_produce: bool = False,
        msg_id: Union[str, None] = None,
    ):
        """Produces a message into a station.
        Args:
            message (bytearray/dict): message to send into the station - bytearray/protobuf class (schema validated station - protobuf) or bytearray/dict (schema validated station - json schema) or string/bytearray/graphql.language.ast.DocumentNode (schema validated station - graphql schema)
            ack_wait_sec (int, optional): max time in seconds to wait for an ack from memphis. Defaults to 15.
            headers (dict, optional): Message headers, defaults to {}.
            async_produce (boolean, optional): produce operation won't wait for broker acknowledgement
            msg_id (string, optional): Attach msg-id header to the message in order to achieve idempotency
        Raises:
            Exception: _description_
        """
        try:
            message = await self.validate_msg(message)

            memphis_headers = {
                "$memphis_producedBy": self.producer_name,
                "$memphis_connectionId": self.connection.connection_id,
            }

            if msg_id is not None and msg_id != "":
                memphis_headers["msg-id"] = msg_id

            if headers is not None:
                headers = headers.headers
                headers.update(memphis_headers)
            else:
                headers = memphis_headers

            if async_produce:
                try:
                    self.loop.create_task(
                        self.connection.broker_connection.publish(
                            self.internal_station_name + ".final",
                            message,
                            timeout=ack_wait_sec,
                            headers=headers,
                        )
                    )
                    await asyncio.sleep(0)
                except Exception as e:
                    raise MemphisError(e)
            else:
                await self.connection.broker_connection.publish(
                    self.internal_station_name + ".final",
                    message,
                    timeout=ack_wait_sec,
                    headers=headers,
                )
        except Exception as e:
            if hasattr(e, "status_code") and e.status_code == "503":
                raise MemphisError(
                    "Produce operation has failed, please check whether Station/Producer still exist"
                )
            else:
                if "Schema validation has failed" in str(
                    e
                ) or "Unsupported message type" in str(e):
                    if self.connection.schema_updates_data[self.internal_station_name] != {}:
                        msg_to_send = ""
                        if hasattr(message, "SerializeToString"):
                            msg_to_send = message.SerializeToString().decode("utf-8")
                        elif isinstance(message, bytearray):
                            msg_to_send = str(message, "utf-8")
                        else:
                            msg_to_send = str(message)
                        if self.connection.station_schemaverse_to_dls[
                            self.internal_station_name
                        ]:
                            unix_time = int(time.time())

                            memphis_headers = {
                                "$memphis_producedBy": self.producer_name,
                                "$memphis_connectionId": self.connection.connection_id,
                            }

                            if headers != {} and not headers == None:
                                headers = headers.headers
                                headers.update(memphis_headers)
                            else:
                                headers = memphis_headers

                            msg_to_send_encoded = msg_to_send.encode("utf-8")
                            msg_hex = msg_to_send_encoded.hex()
                            buf = {
                                "station_name": self.internal_station_name,
                                "producer": {
                                    "name": self.producer_name,
                                    "connection_id": self.connection.connection_id,
                                },
                                "creation_unix": unix_time,
                                "message": {
                                    "data": msg_hex,
                                    "headers": headers,
                                },
                                "validation_error": str(e)
                            }
                            buf = json.dumps(buf).encode("utf-8")
                            await self.connection.broker_manager.publish("$memphis_schemaverse_dls", buf)
                            if self.connection.cluster_configurations.get(
                                "send_notification"
                            ):
                                await self.connection.send_notification(
                                    "Schema validation has failed",
                                    "Station: "
                                    + self.station_name
                                    + "\nProducer: "
                                    + self.producer_name
                                    + "\nError:"
                                    + str(e),
                                    msg_to_send,
                                    schemaverse_fail_alert_type,
                                )
                raise MemphisError(str(e)) from e

    async def destroy(self):
        """Destroy the producer."""
        try:
            destroy_producer_req = {
                "name": self.producer_name,
                "station_name": self.station_name,
                "username": self.connection.username
            }

            producer_name = json.dumps(destroy_producer_req).encode("utf-8")
            res = await self.connection.broker_manager.request(
                "$memphis_producer_destructions", producer_name, timeout=5
            )
            error = res.data.decode("utf-8")
            if error != "" and not "not exist" in error:
                raise Exception(error)

            internal_station_name = get_internal_name(self.station_name)
            producer_number = (
                self.connection.producers_per_station.get(
                    internal_station_name) - 1
            )
            self.connection.producers_per_station[
                internal_station_name
            ] = producer_number

            if producer_number == 0:
                sub = self.connection.schema_updates_subs.get(
                    internal_station_name)
                task = self.connection.schema_tasks.get(internal_station_name)
                if internal_station_name in self.connection.schema_updates_data:
                    del self.connection.schema_updates_data[internal_station_name]
                if internal_station_name in self.connection.schema_updates_subs:
                    del self.connection.schema_updates_subs[internal_station_name]
                if internal_station_name in self.connection.schema_tasks:
                    del self.connection.schema_tasks[internal_station_name]
                if task is not None:
                    task.cancel()
                if sub is not None:
                    await sub.unsubscribe()

            map_key = internal_station_name + "_" + self.real_name
            del self.connection.producers_map[map_key]

        except Exception as e:
            raise Exception(e)
